@@ -19,6 +19,8 @@
 #include <config.h>
 #include <libgnome/libgnome.h>
 #include <signal.h>
+#include <limits.h>
+#include <stdlib.h>
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -361,6 +363,7 @@ gdm_config_parse (void)
     if (ve_string_empty (GdmServAuthDir)) {
 	    gdm_text_message_dialog
 		    (_("No daemon/ServAuthDir specified in the configuration file"));
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: No authdir specified."), "gdm_config_parse");
     }
 
@@ -390,6 +393,8 @@ gdm_config_parse (void)
 			    (GDM_KEY_SERVER_FLEXIBLE);
 		    svr->choosable = gnome_config_get_bool
 			    (GDM_KEY_SERVER_CHOOSABLE);
+		    svr->handled = gnome_config_get_bool
+			    (GDM_KEY_SERVER_HANDLED);
 
 		    if (ve_string_empty (svr->command)) {
 			    gdm_error (_("%s: Empty server command, "
@@ -419,6 +424,7 @@ gdm_config_parse (void)
 	    svr->command = g_strdup (GdmStandardXServer);
 	    svr->flexible = TRUE;
 	    svr->choosable = TRUE;
+	    svr->handled = TRUE;
 
 	    xservers = g_slist_append (xservers, svr);
     }
@@ -499,6 +505,7 @@ gdm_config_parse (void)
 			       "and restart gdm."),
 			     GDM_CONFIG_FILE);
 		    gdm_text_message_dialog (s);
+		    GdmPidFile = NULL;
 		    gdm_fail (_("%s: XDMCP disabled and no local servers defined. Aborting!"), "gdm_config_parse");
 	    }
     }
@@ -520,6 +527,7 @@ gdm_config_parse (void)
 		       "and restart gdm."),
 		     GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: Can't find the gdm user (%s). Aborting!"), "gdm_config_parse", GdmUser);
     } else {
 	    GdmUserId = pwent->pw_uid;
@@ -533,6 +541,7 @@ gdm_config_parse (void)
 		       "correct gdm configuration %s and "
 		       "restart gdm."), GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: The gdm user should not be root. Aborting!"), "gdm_config_parse");
     }
 
@@ -552,6 +561,7 @@ gdm_config_parse (void)
 		       "and restart gdm."),
 		     GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: Can't find the gdm group (%s). Aborting!"), "gdm_config_parse", GdmGroup);
     } else  {
 	    GdmGroupId = grent->gr_gid;   
@@ -565,6 +575,7 @@ gdm_config_parse (void)
 		       "correct gdm configuration %s and "
 		       "restart gdm."), GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: The gdm group should not be root. Aborting!"), "gdm_config_parse");
     }
 
@@ -609,6 +620,7 @@ gdm_config_parse (void)
 		       "restart gdm."), GdmServAuthDir,
 		     GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: Authdir %s does not exist. Aborting."), "gdm_config_parse", GdmServAuthDir);
     }
 
@@ -621,6 +633,7 @@ gdm_config_parse (void)
 		       "restart gdm."), GdmServAuthDir,
 		     GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: Authdir %s is not a directory. Aborting."), "gdm_config_parse", GdmServAuthDir);
     }
 
@@ -635,6 +648,7 @@ gdm_config_parse (void)
 		     GdmServAuthDir, GdmUser, GdmGroup,
 		     GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: Authdir %s is not owned by user %s, group %s. Aborting."), "gdm_config_parse", 
 		      GdmServAuthDir, GdmUser, GdmGroup);
     }
@@ -650,6 +664,7 @@ gdm_config_parse (void)
 		       "restart gdm."),
 		     GdmServAuthDir, GDM_CONFIG_FILE);
 	    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: Authdir %s has wrong permissions %o. Should be 0750. Aborting."), "gdm_config_parse", 
 		      GdmServAuthDir, statbuf.st_mode);
     }
@@ -705,10 +720,11 @@ gdm_daemonify (void)
     }
 
     if (pid < 0) 
-	gdm_fail (_("gdm_daemonify: fork() failed!"));
+	gdm_fail (_("%s: fork() failed!"), "gdm_daemonify");
 
     if (setsid() < 0)
-	gdm_fail (_("gdm_daemonify: setsid() failed: %s!"), strerror(errno));
+	gdm_fail (_("%s: setsid() failed: %s!"), "gdm_daemonify",
+		  strerror(errno));
 
     chdir (GdmServAuthDir);
     umask (022);
@@ -751,6 +767,7 @@ gdm_start_first_unborn_local (int delay)
 				 * autologged in */
 				gdm_first_login = FALSE;
 			} else {
+				gdm_first_login = FALSE;
 				break;
 			}
 		}
@@ -1203,7 +1220,7 @@ gdm_cleanup_children (void)
 static void
 gdm_restart_now (void)
 {
-	gdm_info (_("Gdm restarting ..."));
+	gdm_info (_("GDM restarting ..."));
 	final_cleanup ();
 	if (stored_path != NULL)
 		gnome_setenv ("PATH", stored_path, TRUE);
@@ -1318,12 +1335,41 @@ struct poptOption options [] = {
 	{ NULL, 0, 0, NULL, 0}
 };
 
+static gboolean
+linux_only_is_running (pid_t pid)
+{
+	char resolved_self[PATH_MAX];
+	char resolved_running[PATH_MAX];
+
+	char *running = g_strdup_printf ("/proc/%lu/exe", (gulong)pid);
+
+	if (realpath ("/proc/self/exe", resolved_self) == NULL) {
+		g_free (running);
+		/* probably not a linux system */
+		return FALSE;
+	}
+
+	if (realpath (running, resolved_running) == NULL) {
+		g_free (running);
+		/* probably not a linux system */
+		return FALSE;
+	}
+
+	g_free (running);
+
+	if (strcmp (resolved_running, resolved_self) == 0)
+		return TRUE;
+	return FALSE;
+}
+
 int 
 main (int argc, char *argv[])
 {
     sigset_t mask;
     struct sigaction term, child;
     FILE *pf;
+    poptContext ctx;
+    int nextopt;
 
     store_argv (argc, argv);
 
@@ -1335,12 +1381,19 @@ main (int argc, char *argv[])
     /* Initialize runtime environment */
     umask (022);
 
-    gnome_program_init ("gdm", VERSION,
-			LIBGNOME_MODULE /* module_info */,
-			argc, argv,
-			GNOME_PARAM_POPT_TABLE, options,
-			GNOME_PARAM_CREATE_DIRECTORIES, FALSE,
-			NULL);
+    ctx = poptGetContext ("gdm", argc, (const char **) argv,
+			  options, 0);
+    while ((nextopt = poptGetNextOpt (ctx)) > 0 || nextopt == POPT_ERROR_BADOPT)
+	/* do nothing */ ;
+
+    if (nextopt != -1) {
+	    printf (_("Error on option %s: %s.\nRun '%s --help' to see a full list of available command line options.\n"),
+		    poptBadOption (ctx, 0),
+		    poptStrerror (nextopt),
+		    argv[0]);
+	    fflush (stdout);
+	    exit (1);
+    }
 
     /* XDM compliant error message */
     if (getuid () != 0) {
@@ -1366,7 +1419,8 @@ main (int argc, char *argv[])
 
         if (pf != NULL &&
 	    fscanf (pf, "%d", &pidv) == 1 &&
-	    kill (pidv, 0) != -1) {
+	    kill (pidv, 0) != -1 &&
+	    linux_only_is_running (pidv)) {
 		/* make sure the pid file doesn't get wiped */
 		GdmPidFile = NULL;
 		gdm_fail (_("gdm already running. Aborting!"));
@@ -1894,12 +1948,14 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 
 /* extract second word and the rest of the string */
 static void
-extract_dispname_xauthfile_cookie (const char *msg,
-				   char **dispname,
-				   char **xauthfile,
-				   char **cookie)
+extract_dispname_uid_xauthfile_cookie (const char *msg,
+				       char **dispname,
+				       uid_t *uid,
+				       char **xauthfile,
+				       char **cookie)
 {
 	const char *p;
+	int i;
 	char *pp;
 
 	*dispname = NULL;
@@ -1919,6 +1975,23 @@ extract_dispname_xauthfile_cookie (const char *msg,
 	if (pp != NULL)
 		*pp = '\0';
 
+	/* Get uid */
+	p = strchr (p, ' ');
+	if (p == NULL) {
+		*dispname = NULL;
+		g_free (*dispname);
+		return;
+	}
+	while (*p == ' ')
+		p++;
+
+	if (sscanf (p, "%d", &i) != 1) {
+		*dispname = NULL;
+		g_free (*dispname);
+		return;
+	}
+	*uid = i;
+
 	/* Get cookie */
 	p = strchr (p, ' ');
 	if (p == NULL) {
@@ -2070,7 +2143,9 @@ check_cookie (const char *file, const char *disp, const char *cookie)
 
 static void
 handle_flexi_server (GdmConnection *conn, int type, const char *server,
-		     const char *xnest_disp, const char *xnest_auth_file,
+		     gboolean handled,
+		     const char *xnest_disp, uid_t xnest_uid,
+		     const char *xnest_auth_file,
 		     const char *xnest_cookie)
 {
 	GdmDisplay *display;
@@ -2083,6 +2158,8 @@ handle_flexi_server (GdmConnection *conn, int type, const char *server,
 		struct stat s;
 		gboolean authorized = TRUE;
 
+		seteuid (xnest_uid);
+
 		g_assert (xnest_auth_file != NULL);
 		g_assert (xnest_disp != NULL);
 		g_assert (xnest_cookie != NULL);
@@ -2101,6 +2178,11 @@ handle_flexi_server (GdmConnection *conn, int type, const char *server,
 			authorized = FALSE;
 		}
 
+		if (s.st_uid != xnest_uid)
+			authorized = FALSE;
+
+		seteuid (0);
+
 		if ( ! authorized) {
 			/* Sorry dude, you're not doing something
 			 * right */
@@ -2134,6 +2216,7 @@ handle_flexi_server (GdmConnection *conn, int type, const char *server,
 				      "ERROR 2 Startup errors\n");
 		return;
 	}
+	display->handled = handled;
 
 	if (type == TYPE_FLEXI_XNEST) {
 		GdmDisplay *parent;
@@ -2468,7 +2551,8 @@ gdm_handle_user_message (GdmConnection *conn, const char *msg, gpointer data)
 			return;
 		}
 		handle_flexi_server (conn, TYPE_FLEXI, GdmStandardXServer,
-				     NULL, NULL, NULL);
+				     TRUE /* handled */,
+				     NULL, 0, NULL, NULL);
 	} else if (strncmp (msg, GDM_SUP_FLEXI_XSERVER " ",
 		            strlen (GDM_SUP_FLEXI_XSERVER " ")) == 0) {
 		char *name;
@@ -2509,13 +2593,15 @@ gdm_handle_user_message (GdmConnection *conn, const char *msg, gpointer data)
 		}
 		g_free (name);
 
-		handle_flexi_server (conn, TYPE_FLEXI, command,
-				     NULL, NULL, NULL);
+		handle_flexi_server (conn, TYPE_FLEXI, command, svr->handled,
+				     NULL, 0, NULL, NULL);
 	} else if (strncmp (msg, GDM_SUP_FLEXI_XNEST " ",
 		            strlen (GDM_SUP_FLEXI_XNEST " ")) == 0) {
 		char *dispname = NULL, *xauthfile = NULL, *cookie = NULL;
+		uid_t uid;
 
-		extract_dispname_xauthfile_cookie (msg, &dispname, &xauthfile, &cookie);
+		extract_dispname_uid_xauthfile_cookie (msg, &dispname, &uid,
+						       &xauthfile, &cookie);
 
 		if (dispname == NULL) {
 			/* Something bogus is going on, so just whack the
@@ -2536,7 +2622,8 @@ gdm_handle_user_message (GdmConnection *conn, const char *msg, gpointer data)
 		}
 		
 		handle_flexi_server (conn, TYPE_FLEXI_XNEST, GdmXnest,
-				     dispname, xauthfile, cookie);
+				     TRUE /* handled */,
+				     dispname, uid, xauthfile, cookie);
 
 		g_free (dispname);
 		g_free (xauthfile);
@@ -2575,9 +2662,11 @@ gdm_handle_user_message (GdmConnection *conn, const char *msg, gpointer data)
 		gdm_connection_write (conn, "OK");
 		for (li = displays; li != NULL; li = li->next) {
 			GdmDisplay *disp = li->data;
-			gdm_connection_printf (conn, "%s%ld",
-					       sep, (long)disp->greetpid);
-			sep = ";";
+			if (disp->greetpid > 0) {
+				gdm_connection_printf (conn, "%s%ld",
+						       sep, (long)disp->greetpid);
+				sep = ";";
+			}
 		}
 		gdm_connection_write (conn, "\n");
 	} else if (strncmp (msg, GDM_SUP_UPDATE_CONFIG " ",
